@@ -4,10 +4,10 @@
 Request types definitions
 '''
 
-import six
 import msgpack
 import hashlib
 
+
 from tarantool.const import (
     IPROTO_CODE,
     IPROTO_SYNC,
@@ -24,6 +24,9 @@ from tarantool.const import (
     IPROTO_CLUSTER_UUID,
     IPROTO_VCLOCK,
     IPROTO_EXPR,
+    IPROTO_OPS,
+    # IPROTO_INDEX_BASE,
+    IPROTO_SCHEMA_ID,
     REQUEST_TYPE_OK,
     REQUEST_TYPE_PING,
     REQUEST_TYPE_SELECT,
@@ -31,13 +34,18 @@ from tarantool.const import (
     REQUEST_TYPE_REPLACE,
     REQUEST_TYPE_DELETE,
     REQUEST_TYPE_UPDATE,
+    REQUEST_TYPE_UPSERT,
+    REQUEST_TYPE_CALL16,
     REQUEST_TYPE_CALL,
     REQUEST_TYPE_EVAL,
     REQUEST_TYPE_AUTHENTICATE,
     REQUEST_TYPE_JOIN,
     REQUEST_TYPE_SUBSCRIBE
 )
-
+from tarantool.utils import (
+    strxor,
+    binary_types
+)
 
 class Request(object):
     '''
@@ -55,14 +63,16 @@ class Request(object):
         self._bytes = None
         self.conn = conn
         self._sync = None
+        self._body = ''
 
     def __bytes__(self):
-        return self._bytes
+        return self.header(len(self._body)) + self._body
+
     __str__ = __bytes__
 
     @property
     def sync(self):
-        '''\
+        '''
         :type: int
 
         Required field in the server request.
@@ -73,7 +83,8 @@ class Request(object):
     def header(self, length):
         self._sync = self.conn.generate_sync()
         header = msgpack.dumps({IPROTO_CODE: self.request_type,
-                                IPROTO_SYNC: self._sync})
+                                IPROTO_SYNC: self._sync,
+                                IPROTO_SCHEMA_ID: self.conn.schema_version})
 
         return msgpack.dumps(length + len(header)) + header
 
@@ -94,7 +105,7 @@ class RequestInsert(Request):
         request_body = msgpack.dumps({IPROTO_SPACE_ID: space_no,
                                       IPROTO_TUPLE: values})
 
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestAuthenticate(Request):
@@ -110,22 +121,29 @@ class RequestAuthenticate(Request):
             sha = hashlib.sha1()
             for i in values:
                 if i is not None:
-                    sha.update(i if isinstance(i, six.binary_type) else i.encode())
+                    if isinstance(i, binary_types):
+                        sha.update(i)
+                    else:
+                        sha.update(i.encode())
             return sha.digest()
 
-        def strxor(rhs, lhs):
-            if six.PY2:
-                return "".join(chr(ord(x) ^ ord(y)) for x, y in zip(rhs, lhs))
-
-            return bytes([x ^ y for x, y in zip(rhs, lhs)])
-
         hash1 = sha1((password,))
         hash2 = sha1((hash1,))
         scramble = sha1((salt, hash2))
         scramble = strxor(hash1, scramble)
         request_body = msgpack.dumps({IPROTO_USER_NAME: user,
                                       IPROTO_TUPLE: ("chap-sha1", scramble)})
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
+
+    def header(self, length):
+        self._sync = self.conn.generate_sync()
+        # Set IPROTO_SCHEMA_ID: 0 to avoid SchemaReloadException
+        # It is ok to use 0 in auth every time.
+        header = msgpack.dumps({IPROTO_CODE: self.request_type,
+                                IPROTO_SYNC: self._sync,
+                                IPROTO_SCHEMA_ID: 0})
+
+        return msgpack.dumps(length + len(header)) + header
 
 
 class RequestReplace(Request):
@@ -144,7 +162,7 @@ class RequestReplace(Request):
         request_body = msgpack.dumps({IPROTO_SPACE_ID: space_no,
                                       IPROTO_TUPLE: values})
 
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestDelete(Request):
@@ -163,7 +181,7 @@ class RequestDelete(Request):
                                       IPROTO_INDEX_ID: index_no,
                                       IPROTO_KEY: key})
 
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestSelect(Request):
@@ -182,7 +200,7 @@ class RequestSelect(Request):
                                       IPROTO_ITERATOR: iterator,
                                       IPROTO_KEY: key})
 
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestUpdate(Request):
@@ -201,7 +219,7 @@ class RequestUpdate(Request):
                                       IPROTO_KEY: key,
                                       IPROTO_TUPLE: op_list})
 
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestCall(Request):
@@ -211,14 +229,16 @@ class RequestCall(Request):
     request_type = REQUEST_TYPE_CALL
 
     # pylint: disable=W0231
-    def __init__(self, conn, name, args):
+    def __init__(self, conn, name, args, call_16):
+        if call_16:
+            self.request_type = REQUEST_TYPE_CALL16
         super(RequestCall, self).__init__(conn)
         assert isinstance(args, (list, tuple))
 
         request_body = msgpack.dumps({IPROTO_FUNCTION_NAME: name,
                                       IPROTO_TUPLE: args})
 
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestEval(Request):
@@ -235,7 +255,7 @@ class RequestEval(Request):
         request_body = msgpack.dumps({IPROTO_EXPR: name,
                                       IPROTO_TUPLE: args})
 
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestPing(Request):
@@ -246,7 +266,26 @@ class RequestPing(Request):
 
     def __init__(self, conn):
         super(RequestPing, self).__init__(conn)
-        self._bytes = self.header(0)
+        self._body = b''
+
+
+class RequestUpsert(Request):
+    '''
+    Represents UPSERT request
+    '''
+
+    request_type = REQUEST_TYPE_UPSERT
+
+    # pylint: disable=W0231
+    def __init__(self, conn, space_no, index_no, tuple_value, op_list):
+        super(RequestUpsert, self).__init__(conn)
+
+        request_body = msgpack.dumps({IPROTO_SPACE_ID: space_no,
+                                      IPROTO_INDEX_ID: index_no,
+                                      IPROTO_TUPLE: tuple_value,
+                                      IPROTO_OPS: op_list})
+
+        self._body = request_body
 
 
 class RequestJoin(Request):
@@ -259,7 +298,7 @@ class RequestJoin(Request):
     def __init__(self, conn, server_uuid):
         super(RequestJoin, self).__init__(conn)
         request_body = msgpack.dumps({IPROTO_SERVER_UUID: server_uuid})
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
 
 
 class RequestSubscribe(Request):
@@ -278,7 +317,8 @@ class RequestSubscribe(Request):
             IPROTO_SERVER_UUID: server_uuid,
             IPROTO_VCLOCK: vclock
         })
-        self._bytes = self.header(len(request_body)) + request_body
+        self._body = request_body
+
 
 class RequestOK(Request):
     '''
@@ -289,6 +329,6 @@ class RequestOK(Request):
     # pylint: disable=W0231
     def __init__(self, conn, sync):
         super(RequestOK, self).__init__(conn)
-        header = msgpack.dumps({IPROTO_CODE: self.request_type,
-                                IPROTO_SYNC: sync})
-        self._bytes = msgpack.dumps(len(header)) + header
+        request_body = msgpack.dumps({IPROTO_CODE: self.request_type,
+                                      IPROTO_SYNC: sync})
+        self._body = request_body
